from __future__ import absolute_import, print_function

import hashlib
import inspect
import logging
import os
import sys
from abc import abstractmethod

import pandas as pd
import pandas.io.sql
import six
import sqlparse
from decorator import decorator
from jinja2 import StrictUndefined, Template

from . import cursor_formatters
from omniduct.caches.base import cached_method
from omniduct.duct import Duct
from omniduct.utils.config import config
from omniduct.utils.debug import logger, logging_scope
from omniduct.utils.magics import MagicsProvider, process_line_arguments, process_line_cell_arguments

logging.getLogger('requests').setLevel(logging.WARNING)


@decorator
def render_statement(method, self, statement, *args, **kwargs):
    # Check if statement is an SQLAlchemy executable expression, and if so, render it
    try:
        from sqlalchemy.sql.base import Executable
        if isinstance(statement, Executable):
            statement = str(statement.compile(compile_kwargs={"literal_binds": True}))
    except ImportError:
        pass

    # If templating enabled, render template
    template = kwargs.pop('template', False)
    template_context = kwargs.pop('template_context', {})

    if template:
        statement = self.render_template(statement, template_context)

    return method(self, statement, *args, **kwargs)


class DatabaseClient(Duct, MagicsProvider):
    '''
    QueryClient is an abstract class that can be subclassed to allow the databases
    of various data sources, such as databases or website apis.
    '''

    DUCT_TYPE = Duct.Type.DATABASE
    DEFAULT_PORT = None

    CURSOR_FORMATTERS = {
        'pandas': cursor_formatters.PandasCursorFormatter,
        'hive': cursor_formatters.HiveCursorFormatter,
        'csv': cursor_formatters.CsvCursorFormatter,
        'tuple': cursor_formatters.TupleCursorFormatter,
        'dict': cursor_formatters.DictCursorFormatter,
        'raw': cursor_formatters.RawCursorFormatter,
    }
    DEFAULT_CURSOR_FORMATTER = 'pandas'

    def __init__(self, *args, **kwargs):
        '''
        This is a shim __init__ function that passes all arguments onto
        `self._init`, which is implemented by subclasses. This allows subclasses
        to instantiate themselves with arbitrary parameters.
        '''
        Duct.__init_with_kwargs__(self, kwargs, port=self.DEFAULT_PORT)

        self._templates = {}
        self._sqlalchemy_engine = None
        self._sqlalchemy_metadata = None

        self._init(*args, **kwargs)

    @abstractmethod
    def _init(self):
        pass

    def __call__(self, query, **kwargs):
        """Calls DatabaseClient.query() largely for backwards compatibility"""
        return self.query(query, **kwargs)

    # Querying
    @classmethod
    def statements_split(cls, statements):
        for statement in sqlparse.split(statements):
            statement = statement.strip()
            if statement.endswith(';'):
                statement = statement[:-1].strip()
            if statement:  # remove empty statements
                yield statement

    @classmethod
    def statement_cleanup(cls, statement):
        statement = sqlparse.format(statement, strip_comments=True, reindent=True)
        statement = os.linesep.join([line for line in statement.splitlines() if line])
        return statement

    @classmethod
    def statement_hash(cls, statement, cleanup=True):
        if cleanup:
            statement = cls.statement_cleanup(statement)
        if sys.version_info.major == 3 or sys.version_info.major == 2 and isinstance(statement, unicode):
            statement = statement.encode('utf8')
        return hashlib.sha256(statement).hexdigest()

    @render_statement
    def execute(self, statement, cleanup=True, async=False, cursor=None, **kwargs):
        '''
        Execute a statement against the data source.

        Parameters
        ----------
        statement : The statement to be executed by the query client (possibly templated).
        cleanup : Whether statement should be normalised (whitespace and comments removed).
                  This helps to avoid missing the cache for similar queries.
        async : Whether the cursor should be returned before the server-side query
                computation is complete.
        template : Whether the statement should be treated as a Jinja2 template.
        template_context : If the statement is to be treated as a template,
                 substitute in the parameters using this context.
        cursor : Rather than creating a new cursor, execute this statement against
                 the existing provided cursor (or pass None)
        kwargs : Extra keyword arguments to be passed on to _execute, as implemented by subclasses.

        Returns
        -------
        A DBAPI2 compatible cursor instance.
        '''

        self.connect()

        statements = self.statements_split(statement)
        statements = [self.statement_cleanup(stmt) if cleanup else stmt for stmt in statements]
        assert len(statements) > 0, "No non-empty statements were provided."

        for statement in statements[:-1]:
            cursor = self.connect()._execute(statement, cursor=cursor, async=False, **kwargs)
        cursor = self.connect()._execute(statements[-1], cursor=cursor, async=async, **kwargs)

        return cursor

    @logging_scope("Query", timed=True)
    @render_statement
    @cached_method(
        id_str=lambda self, kwargs: "{}:\n{}".format(kwargs['format'], self.statement_hash(kwargs['statement'], cleanup=kwargs.get('cleanup', True))),
    )
    def query(self, statement, format=None, format_opts={}, **kwargs):
        '''
        Execute a statement against the data source using `.execute()`, and then
        collect and return the results in the nominated format; optionally (and
        by default) caching the result.

        Parameters
        ----------
        statement : The statement to be executed by the query client (possibly templated).
        format : A subclass of CursorFormatter, or one of: 'pandas', 'hive',
                'csv', 'tuple' or 'dict'.
        format_opts : A dictionary of format-specific options.
        use_cache : True (default) or False. Whether to use the cache (if present).
        renew : True or False (default). If cache is being used, renew it before
                returning stored value.
        **kwargs : Additional arguments to pass on to `.execute()`.
        '''
        cursor = self.execute(statement, async=False, template=False, **kwargs)

        # Some DBAPI2 cursor implementations error if attempting to extract
        # data from an empty cursor, and if so, we simply return None.
        if self._cursor_empty(cursor):
            return None

        formatter = self._get_formatter(format, cursor, **format_opts)
        return formatter.dump()

    def stream(self, statement, format=None, format_opts={}, batch=None, **kwargs):
        cursor = self.execute(statement, async=False, **kwargs)
        formatter = self._get_formatter(format, cursor, **format_opts)

        for row in formatter.stream(batch=batch):
            yield row

    def _get_formatter(self, formatter, cursor, **kwargs):
        formatter = formatter or self.DEFAULT_CURSOR_FORMATTER
        if not (inspect.isclass(formatter) and issubclass(formatter, cursor_formatters.CursorFormatter)):
            assert formatter in self.CURSOR_FORMATTERS, "Invalid format '{}'. Choose from: {}".format(formatter, ','.join(self.CURSOR_FORMATTERS.keys()))
            formatter = self.CURSOR_FORMATTERS[formatter]
        return formatter(cursor, **kwargs)

    def stream_to_file(self, statement, file, format='csv', **kwargs):
        close_later = False
        if isinstance(file, str):
            file = open(file, 'w')
            close_later = True

        try:
            file.writelines(self.stream(statement, format=format, **kwargs))
        finally:
            if close_later:
                file.close()

    def execute_from_file(self, file, **kwargs):
        """
        Read file contents and execute them against the connected data source.

        Parameters
        ----------
        file : str
            File containing a query
        kwargs : dict
            Extra parameters to pass on to `execute`.
        """
        with open(file, 'r') as f:
            return self.execute(f.read(), **kwargs)

    def query_from_file(self, file, **kwargs):
        '''
        This method is shorthand for:
        QueryClient.execute_from_file(file, parse=True, **kwargs)
        '''
        with open(file, 'r') as f:
            return self.query(f.read(), **kwargs)

    def add_template(self, name, body):
        self._templates[name] = body
        return self

    def render_template(self, name_or_statement, context=None, by_name=False):

        if by_name:
            if name_or_statement not in self._templates:
                raise ValueError("No such template of name: '{}'.".format(name_or_statement))
            statement = self._templates[name_or_statement]
        else:
            statement = name_or_statement

        if not context:
            context = {}

        # Substitute in any other named statements recursively
        while '{{{' in statement or '{{%' in statement:
            statement = Template(statement,
                                 block_start_string='{{%',
                                 block_end_string='%}}',
                                 variable_start_string='{{{',
                                 variable_end_string='}}}',
                                 comment_start_string='{{#',
                                 comment_end_string='#}}',
                                 undefined=StrictUndefined).render(getattr(self, '_templates', {}))

        # Evaluate final template in provided context
        statement = Template(statement, undefined=StrictUndefined).render(context)

        return statement

    def execute_from_template(self, name, context=None, **kwargs):
        statement = self.render_template(name, context, by_name=True)
        return self.execute(statement, **kwargs)

    def query_from_template(self, name, context=None, **kwargs):
        statement = self.render_template(name, context, by_name=True)
        return self.query(statement, **kwargs)

    # Uploading data to data store
    @logging_scope('Push', timed=True)
    def push(self, df, table, if_exists='fail', **kwargs):
        '''
        This method pushes a local dataframe `df` to the connected data store
        as a table `table`.

        Parameters
        ----------
        df : pandas.DataFrame
            The pandas dataframe to push into the data store.
        table : str
            The name of the table into which the dataframe should be pushed.
        if_exists : {'fail', 'replace', 'append'}, default 'fail'
            - fail: If table exists, do nothing.
            - replace: If table exists, drop it, recreate it, and insert data.
            - append: If table exists, insert data. Create if does not exist.
        kwargs : dict
            Additional arguments which are passed on to `QueryClient._push`.
        '''
        assert if_exists in {'fail', 'replace', 'append'}
        self.connect()._push(df, table, if_exists=if_exists, **kwargs)

    # Table properties

    @abstractmethod
    def _execute(self, statement, cursor=None, async=False, **kwargs):
        pass

    def _push(self, df, table, if_exists='fail', **kwargs):
        if self._sqlalchemy_engine is None:
            raise NotImplementedError("Support for pushing data tables using `{}` is not currently implemented.".format(self.__class__.__name__))
        return df.to_sql(name=table, con=self._sqlalchemy_engine, index=False, if_exists=if_exists, **kwargs)

    def _cursor_empty(self, cursor):
        pass

    def _cursor_wait(self, cursor, poll_interval=1):
        pass

    def table_list(self, **kwargs):
        '''
        Return a list of table names in the data source as a DataFrame. Additional kwargs are
        passed to `QueryClient._table_list`.
        '''
        return self._table_list(**kwargs)

    @abstractmethod
    def _table_list(self, **kwargs):
        pass

    def table_exists(self, table, **kwargs):
        '''
        Return boolean if table exists in schema
        '''
        return self._table_exists(table=table, **kwargs)

    @abstractmethod
    def _table_exists(self, table, **kwargs):
        pass

    def table_desc(self, table, **kwargs):
        '''
        Describe a table in the data source. Additional kwargs are
        passed to `QueryClient._table_desc`.

        Returns a pandas dataframe of table fields and descriptors.
        '''
        return self._table_desc(table, **kwargs)

    @abstractmethod
    def _table_desc(self, table, **kwargs):
        pass

    def table_head(self, table, n=10, **kwargs):
        '''
        Show a sample of the data in `table` of the data source. `n` is the number of records
        to show in this sample. The additional `kwargs` are passed on to `QueryClient._table_head`.

        Returns a pandas DataFrame.
        '''
        return self._table_head(table, n=n, **kwargs)

    @abstractmethod
    def _table_head(self, table, n=10, **kwargs):
        pass

    def table_props(self, table, **kwargs):
        '''
        Return a dataframe of table properties for `table`.
        '''
        return self._table_props(table, **kwargs)

    @abstractmethod
    def _table_props(self, table, **kwargs):
        pass

    def _register_magics(self, base_name):
        from IPython.core.magic import register_line_magic, register_cell_magic, register_line_cell_magic

        def statement_executor_magic(executor, statement, variable=None, show='head', transpose=False, template=True, template_context=None, **kwargs):

            ip = get_ipython()

            if template_context is None:
                template_context = ip.user_ns

            # Line magic
            if statement is None:
                return self.query_from_template(variable, context=template_context)

            # Cell magic
            result = getattr(self, executor)(statement, template=template, template_context=template_context, **kwargs)

            if variable is not None:
                ip.user_ns[variable] = result

            if executor != 'query':
                if variable is None:
                    return result
                return
            elif variable is None:
                return result

            format = kwargs.get('format', self.DEFAULT_CURSOR_FORMATTER)
            if show == 'head':
                show = 10
            if isinstance(show, int):
                r = result.head(show) if format == 'pandas' else result[:show]
            elif show == 'all':
                r = result
            elif show == 'none':
                return None
            else:
                raise ValueError("Omniduct does not recognise the argument show='{0}' in cell magic.".format(show))

            if format == 'pandas' and transpose:
                return r.T
            return r

        @register_line_cell_magic(base_name)
        @process_line_cell_arguments
        def query_magic(*args, **kwargs):
            return statement_executor_magic('query', *args, **kwargs)

        @register_line_cell_magic("{}.{}".format(base_name, 'execute'))
        @process_line_cell_arguments
        def query_magic(*args, **kwargs):
            return statement_executor_magic('execute', *args, **kwargs)

        @register_line_cell_magic("{}.{}".format(base_name, 'stream'))
        @process_line_cell_arguments
        def query_magic(*args, **kwargs):
            return statement_executor_magic('stream', *args, **kwargs)

        @register_cell_magic("{}.{}".format(base_name, 'template'))
        @process_line_arguments
        def add_template(body, name):
            self.add_template(name, body)

        @register_line_cell_magic("{}.{}".format(base_name, 'render'))
        @process_line_cell_arguments
        def render_template(body=None, name=None, context=None, show=True):

            ip = get_ipython()

            try:
                if body is None:
                    assert name is not None, "Name must be specified in line-mode."
                    rendered = self.render_template(name, context=context or ip.user_ns, by_name=True)
                    if not show:
                        return rendered
                else:
                    rendered = self.render_template(body, context=context or ip.user_ns, by_name=False)
                    if name is not None:
                        ip.user_ns[name] = rendered
            finally:
                if show:
                    print(rendered)

        @register_line_magic("{}.{}".format(base_name, 'desc'))
        @process_line_arguments
        def table_desc(table_name, **kwargs):
            return self.table_desc(table_name, **kwargs)

        @register_line_magic("{}.{}".format(base_name, 'head'))
        @process_line_arguments
        def table_head(table_name, **kwargs):
            return self.table_head(table_name, **kwargs)

        @register_line_magic("{}.{}".format(base_name, 'props'))
        @process_line_arguments
        def table_props(table_name, **kwargs):
            return self.table_props(table_name, **kwargs)



from __future__ import absolute_import

import logging
import os
import re
import tempfile
import time

import pandas as pd
from jinja2 import Template

from omniduct.utils.config import config
from omniduct.utils.debug import logger
from omniduct.utils.processes import Timeout, run_in_subprocess

from .base import DatabaseClient


class HiveServer2Client(DatabaseClient):

    PROTOCOLS = ['hiveserver2']
    DEFAULT_PORT = 3623

    def _init(self, schema=None, driver='pyhive', auth_mechanism='NOSASL', **connection_options):
        self.schema = schema
        self.driver = driver
        self.auth_mechanism = auth_mechanism
        self.connection_options = connection_options
        self.__hive = None
        self.connection_fields += ('schema',)

        assert self.driver in ('pyhive', 'impyla'), "Supported drivers are pyhive and impyla."

    def _connect(self):
        from sqlalchemy import create_engine, MetaData
        if self.driver == 'pyhive':
            import pyhive.hive
            self.__hive = pyhive.hive.connect(host=self.host,
                                              port=self.port,
                                              auth=self.auth_mechanism,
                                              database=self.schema,
                                              **self.connection_options)
            self._sqlalchemy_engine = create_engine('hive://{}:{}/{}'.format(self.host, self.port, self.schema))
            self._sqlalchemy_metadata = MetaData(self._sqlalchemy_engine)
        elif self.driver == 'impyla':
            import impala.dbapi
            self.__hive = impala.dbapi.connect(host=self.host,
                                               port=self.port,
                                               auth_mechanism=self.auth_mechanism,
                                               database=self.schema,
                                               **self.connection_options)
            self._sqlalchemy_engine = create_engine('impala://{}:{}/{}'.format(self.host, self.port, self.schema))
            self._sqlalchemy_metadata = MetaData(self._sqlalchemy_engine)

    def __hive_cursor(self):
        if self.driver == 'impyla':  # Impyla seems to have all manner of connection issues, attempt to restore connection
            try:
                with Timeout(1):
                    return self.__hive.cursor()
            except:
                self._connect()
        return self.__hive.cursor()

    def _is_connected(self):
        return self.__hive is not None

    def _disconnect(self):
        logger.info('Disconnecting from Hive coordinator...')
        try:
            self.__hive.close()
        except:
            pass
        self.__hive = None
        self._sqlalchemy_engine = None
        self._sqlalchemy_metadata = None

    def _execute(self, statement, cursor=None, poll_interval=1, async=False):
        """
        Execute command

        poll_interval : int, optional
            Default delay in polling for query status
        """
        cursor = cursor or self.__hive_cursor()
        log_offset = 0

        if self.driver == 'pyhive':
            from TCLIService.ttypes import TOperationState
            cursor.execute(statement, async=True)

            if not async:
                status = cursor.poll().operationState
                while status in (TOperationState.INITIALIZED_STATE, TOperationState.RUNNING_STATE):
                    log_offset = self._log_status(cursor, log_offset)
                    time.sleep(poll_interval)
                    status = cursor.poll().operationState

        elif self.driver == 'impyla':
            cursor.execute_async(statement)
            if not async:
                while cursor.is_executing():
                    log_offset = self._log_status(cursor, log_offset)
                    time.sleep(poll_interval)

        return cursor

    def _cursor_empty(self, cursor):
        if self.driver == 'impyla':
            return not cursor.has_result_set
        elif self.driver == 'pyhive':
            return cursor.description is None
        return False

    def _cursor_wait(self, cursor, poll_interval=1):
        status = cursor.poll().operationState
        while status in (TOperationState.INITIALIZED_STATE, TOperationState.RUNNING_STATE):
            time.sleep(poll_interval)
            status = cursor.poll().operationState

    def _log_status(self, cursor, log_offset=0):
        matcher = re.compile('[0-9/]+ [0-9\:]+ (INFO )?')

        if self.driver == 'pyhive':
            log = cursor.fetch_logs()
        else:
            log = cursor.get_log().strip().split('\n')

        for line in log[log_offset:]:
            if not line:
                continue
            m = matcher.match(line)
            if m:
                line = line[len(m.group(0)):]
            logger.info(line)

        return len(log)

    def _push(self, df, table, if_exists='fail', schema=None, **kwargs):
        try:
            return DatabaseClient._push(self, df, table, if_exists=if_exists, schema=schema or self.username, **kwargs)
        except Exception as e:
            raise RuntimeError("Push unsuccessful. Your version of Hive may be too old to support the `INSERT` keyword. Original exception was: {}".format(e.args[0]))

    def _table_list(self, schema=None, like='*', **kwargs):
        schema = schema or self.schema or 'default'
        return self.query("SHOW TABLES IN {0} '{1}'".format(schema, like),
                          **kwargs)

    def _table_exists(self, table, schema=None):
        return (self.table_list(renew=True, schema=schema)['tab_name'] == table).any()

    def _table_desc(self, table, **kwargs):
        records = self.query("DESCRIBE {0}".format(table), **kwargs)

        # pretty hacky but hive doesn't return DESCRIBE results in a nice format
        # TODO is there any information we should pull out of DESCRIBE EXTENDED
        for i, record in enumerate(records):
            if record[0] == '':
                break

        columns = ['col_name', 'data_type', 'comment']
        fields_df = pd.DataFrame(records[:i], columns=columns)

        partitions_df = pd.DataFrame(records[i + 4:], columns=columns)
        partitions_df['comment'] = "PARTITION " + partitions_df['comment']

        return pd.concat((fields_df, partitions_df))

    def _table_head(self, table, n=10, **kwargs):
        return self.query("SELECT * FROM {} LIMIT {}".format(table, n), **kwargs)

    def _table_props(self, table, **kwargs):
        return self.query('SHOW TBLPROPERTIES `{0}`'.format(table), **kwargs)

    def _run_in_hivecli(self, cmd):
        """Run a query using hive cli in a subprocess."""
        # Turn hive command into quotable string.
        double_escaped = re.sub('\\' * 2, '\\' * 4, cmd)
        sys_cmd = 'hive -e "{0}"'.format(re.sub('"', '\\"', double_escaped))
        # Execute command in a subprocess.
        if self.remote:
            proc = self.remote.execute(sys_cmd)
        else:
            proc = run_in_subprocess(sys_cmd, check_output=True)
        return proc


def _create_table_statement_from_df(df, table, schema='default', drop=False,
                                    text=True, sep=None, loc=None):
    """
    Return create table statement for new hive table based on pandas dataframe.

    Parameters
    ----------
    df : pandas.DataFrame or pandas.Series
        Used to determine column names and types for create table statement.
    table : str
        Table name for create table statement.
    schema : str
        Schema for create table statement
    drop : bool
        Whether to include a drop table statement along with create table statement.
    text : bool
        Whether data will be stored as a text file.
    sep : str
        Field delimiter for text file (only used if text==True).
    loc : str, optional
        Desired hdfs location.

    Returns
    -------
    cmd : str
        A create table statement.
    """
    # dtype kind to hive type mapping dict.
    DTYPE_KIND_HIVE_TYPE = {
        'b': 'BOOLEAN',  # boolean
        'i': 'BIGINT',   # signed integer
        'u': 'BIGINT',   # unsigned integer
        'f': 'DOUBLE',   # floating-point
        'c': 'STRING',   # complex floating-point
        'O': 'STRING',   # object
        'S': 'STRING',   # (byte-)string
        'U': 'STRING',   # Unicode
        'V': 'STRING'    # void
    }
    sep = sep or "\t"

    # Sanitive column names and map data types to hive types.
    columns = []
    for col, dtype in df.dtypes.iteritems():
        col_sanitized = re.sub('\W', '', col.lower().replace(' ', '_'))
        hive_type = DTYPE_KIND_HIVE_TYPE[dtype.kind]
        columns.append('  {column}  {type}'.format(column=col_sanitized,
                                                   type=hive_type))

    cmd = Template("""
    {% if drop %}
    DROP TABLE IF EXISTS {{ schema }}.{{ table }};
    {% endif -%}
    CREATE TABLE IF NOT EXISTS {{ schema }}.{{ table }} (
    {%- for col in columns %}
     {{ col }} {% if not loop.last %}, {% endif %}
    {%- endfor %}
    )
    {%- if text %}
    ROW FORMAT DELIMITED
    FIELDS TERMINATED BY "{{ sep }}"
    STORED AS TEXTFILE
    {% endif %}
    {%- if loc %}
    LOCATION "{{ loc }}"
    {%- endif %}
    ;
    """).render(drop=drop, table=table, schema=schema, columns=columns, text=text, sep=sep)

    logger.debug('Create Table Statement: {}'.format(cmd))
    return cmd



from .caches.local import LocalCache
from .databases.hiveserver2 import HiveServer2Client
from .databases.presto import PrestoClient
from .databases.sqlalchemy import SQLAlchemyClient
from .databases.neo4j import Neo4jClient
from .filesystems.local import LocalFsClient
from .filesystems.webhdfs import WebHdfsClient
from .remotes.ssh import SSHClient
from .restful.base import RestClient

# from .remotes.ssh_paramiko import ParamikoSSHClient  # Not yet ready for prime time



import select
import threading

from omniduct.remotes.base import RemoteClient
from omniduct.utils.debug import logger

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer


class ParamikoSSHClient(RemoteClient):

    PROTOCOLS = ['ssh_paramiko']

    def _init(self):
        logger.warning("The Paramiko SSH client is still under development, \
                        and is not ready for use as a daily driver.")

    def _connect(self):
        import paramiko  # Imported here due to relatively slow import
        self.__client = paramiko.SSHClient()
        self.__client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy())
        self.__client.load_system_host_keys()

        self.__client.connect(self.host, username=self.username)

    def _is_connected(self):
        try:
            return self.__client.get_transport().is_active()
        except:
            return False

    def _disconnect(self):
        try:
            return self.__client.close()
        except:
            pass

    def _execute(self, cmd, **kwargs):
        return self.__client.exec_command(cmd)

    def _copy_to_local(self, source, dest):
        raise NotImplementedError

    def _copy_from_local(self, source, dest):
        raise NotImplementedError

    def _port_forward_start(self, local_port, remote_host, remote_port):
        logger.debug('Now forwarding port {} to {}:{} ...'.format(local_port, remote_host, remote_port))

        remote_host, remote_port = get_host_port(remote_host, remote_port)

        try:
            server = forward_tunnel(local_port, remote_host, remote_port, self.__client.get_transport())
        except KeyboardInterrupt:
            print('C-c: Port forwarding stopped.')
        return server

    def _port_forward_stop(self, local_port, remote_host, remote_port, server):
        server.shutdown()

    def _is_port_bound(self, host, port):
        return True

    # FileSystem methods

    def _exists(self, path):
        raise NotImplementedError

    def _isdir(self, path):
        raise NotImplementedError

    def _isfile(self, path):
        raise NotImplementedError

    def _listdir(self, path):
        raise NotImplementedError

    def _showdir(self, path):
        raise NotImplementedError

    # File handling

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        raise NotImplementedError

    def _file_append_(self, path, s, binary):
        raise NotImplementedError

    def _file_write_(self, path, s, binary):
        raise NotImplementedError


# Port Forwarding Utility Code
# Largely based on code from: https://github.com/paramiko/paramiko/blob/master/demos/forward.py

SSH_PORT = 22
DEFAULT_PORT = 4000

g_verbose = True


class ForwardServer (SocketServer.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler (SocketServer.BaseRequestHandler):

    def handle(self):
        try:
            chan = self.ssh_transport.open_channel('direct-tcpip',
                                                   (self.chain_host, self.chain_port),
                                                   self.request.getpeername())
        except Exception as e:
            logger.info('Incoming request to %s:%d failed: %s' % (self.chain_host,
                                                                  self.chain_port,
                                                                  repr(e)))
            return
        if chan is None:
            logger.info('Incoming request to %s:%d was rejected by the SSH server.' %
                        (self.chain_host, self.chain_port))
            return

        logger.info('Connected!  Tunnel open %r -> %r -> %r' % (self.request.getpeername(),
                                                                chan.getpeername(), (self.chain_host, self.chain_port)))
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        logger.info('Tunnel closed from %r' % (peername,))


def forward_tunnel(local_port, remote_host, remote_port, transport):
    # this is a little convoluted, but lets me configure things for the Handler
    # object.  (SocketServer doesn't give Handlers any way to access the outer
    # server normally.)
    class SubHander (Handler):
        chain_host = remote_host
        chain_port = remote_port
        ssh_transport = transport
    server = ForwardServer(('', local_port), SubHander)

    t = threading.Thread(target=server.serve_forever)
    t.setDaemon(True)  # don't hang on exit
    t.start()

    return server


def get_host_port(spec, default_port=22):
    "parse 'hostname:22' into a host and port, with the port optional"
    args = (spec.split(':', 1) + [default_port])[:2]
    args[1] = int(args[1])
    return args[0], args[1]



import inspect
import io
import json
import logging
import os
import sys

import six


def ensure_path_exists(path):
    path = os.path.expanduser(path)
    if not os.path.exists(path):
        os.makedirs(path)
    return path


logger = logging.getLogger(__name__)


class ConfigurationRegistry(object):

    def __init__(self):
        self._register = {}

    def register(self, key, description=None, default=None, onchange=None, onload=None,
                 type=None, host=None):
        '''
        Register a configuration key that can be set by the user. As noted in the
        class level documentation, these keys should not lead to changes in the
        output of omniduct functions. The same code should generate the same results
        independent of this configuration.

        The arguments to this method are:
         - key : A string used to identify the configuration option.
         - description* : A string description of the configuration key.
         - default* : A default value for the key.
         - onchange* : A function to call when the parameter changes (should have
            a signature accepting one variable).
         - onload* : A function of no arguments to call to initialize the value
            of this configuration setting
         - type : Values set will be of `isinstance(val, type)`

        * If not specified, these fields default to None.
        '''
        if key in dir(self):
            raise KeyError("Key `{0}` cannot be registered as it conflicts with a method of OmniductConfiguration.".format(key))
        if key in self._register:
            logger.warn("Overwriting existing key `{0}`, previously registered by {1}".format(key, self._register[key]['host']))

        try:
            caller_frame = inspect.current_frame().f_back
            host = inspect.getmodule(caller_frame).__name__
        except:
            host = 'unknown'

        if default is not None and type is not None:
            assert isinstance(default, type)
        self._register[key] = {
            'description': description,
            'host': host,
            'default': default,
            'onchange': onchange,
            'onload': onload,
            'type': type,
        }

    def show(self):
        '''
        Pretty print the configuration options available to be set, as well as
        their current values, descriptions and the module from which they were
        registered.
        '''
        for key in sorted(self._register.keys()):
            desc = self._register[key].get('description')
            if desc is None:
                desc = 'No description'
            print('{0} with default = {1}'.format(key, self._register[key]['default']))
            print('\t{0}'.format(desc))
            print('\t({0})'.format(self._register[key]['host']))


class Configuration(ConfigurationRegistry):
    """
    Configuration is a hub for storing runtime configuration settings, as
    well as persisting them to disk. Ideally, it should store only configuration
    options that allow omniduct to run optimally on the system (such as preferred
    hostnames, servers and usernames). In particular, the same code run on different
    systems should output the same results (independent of configuration); so,
    for example, this configuration hub should *not* be used for things like
    default plot styling, etc.

    Retrieving a configuration option looks like:
    >>> config.logging_level
    20

    Setting a configuration option looks like:
    >>> config.logging_level = 10

    Reviewing all available options and set values looks like:
    >>> config.show()
    """

    def __init__(self, *registries, **kwargs):
        ConfigurationRegistry.__init__(self)

        for registry in registries:
            for key, props in registry.items():
                self.register(key, **props)

        self._config = {}
        self.__config_path = kwargs.pop('config_path', None)

    def __dir__(self):
        return sorted(self._register.keys())

    @property
    def _config_path(self):
        return self.__config_path

    @_config_path.setter
    def _config_path(self, path):
        self.__config_path = path

        if path is not None and os.path.exists(self.__config_path):
            # Restore configuration
            try:
                self.load(force=True)
            except:
                raise RuntimeError(
                    "Configuration file at {0} cannot be loaded. Perhaps try deleting it.".format(self.__config_path))

    def all(self):
        '''
        Return a dictionary containing all configuration keys. Note that this is
        the actual dictionary storing the configuration options, so modifying
        this dictionary will modify the configuration options *without* running
        the standard checks.
        '''
        return self._config

    def show(self):
        '''
        Pretty print the configuration options available to be set, as well as
        their current values, descriptions and the module from which they were
        registered.
        '''
        for key in sorted(self._register.keys()):
            desc = self._register[key].get('description')
            if desc is None:
                desc = 'No description'
            val = str(self._config.get(key, '<Not Set>'))
            print('{0} = {1} (default = {2})'.format(key, val, self._register[key]['default']))
            print('\t{0}'.format(desc))
            print('\t({0})'.format(self._register[key]['host']))

    def __setattr__(self, key, value):
        '''
        Allow setting configuration options using the standard python attribute
        methods, as described in the class documentation.

        Attributes prefixed with '_' are loaded from this class.
        '''
        if key.startswith('_'):
            object.__setattr__(self, key, value)
        elif key in self._register:
            if self._register[key]['type'] is not None:
                if not isinstance(value, self._register[key]['type']):
                    raise ValueError(
                        "{} must be in type(s) {}".format(key, self._register[key]['type']))
            if self._register[key]['onchange'] is not None:
                self._register[key]['onchange'](value)
            self._config[key] = value
        else:
            raise KeyError("No such configuration key `{0}`.".format(key))

    def __getattr__(self, key):
        '''
        Allow retrieval of configuration keys using standard python atribute
        methods, as described in the class documentation.

        Attributes prefixed with '_' are loaded from this class.
        '''
        if key.startswith('_'):
            return object.__getattr__(self, key)
        if key in self._register:
            if key in self._config:
                return self._config[key]

            # if a lazy loader is specified, use it
            if self._register[key]['default'] is None and self._register[key]['onload'] is not None:
                setattr(self, key, self._register[key]['onload']())

            return self._config.get(key, self._register[key]['default'])
        raise AttributeError("No such configuration key `{0}`.".format(key))

    def reset(self, *keys, **target_config):
        '''
        Reset all configuration keys specifed to their default values, or values
        specified in `target_config`. If both `keys` and `target_config` are
        specified, `keys` acts to both filter the keys of `target_config` and add
        default values as the missing keys.
        >>> config.reset('logging_level')
        >>> config.reset('logging_level', logging_level=10)
        >>> config.reset(logging_level=10)

        If no keys are specified, reset all keys:
        >>> config.reset()
        '''
        if len(keys) == 0:
            keys = set(list(self._register.keys()) + list(target_config.keys()))

        target_config = self.__restrict_keys(target_config, keys)
        reset_keys = [key for key in keys if key not in target_config]

        for key, value in target_config.items():
            self._config[key] = value
            if key in self._register:
                if value == self._register[key]['default']:
                    self._config.pop(key)
                if self._register[key]['onchange'] is not None:
                    self._register[key]['onchange'](getattr(self, key))
            else:  # Allow users to delete deprecated keys
                logger.warning("Added value for configuration key `{0}` which has yet to be registered.".format(key))

        for key in reset_keys:
            if key in self._config:
                self._config.pop(key)
                if key in self._register and self._register[key]['onchange'] is not None:
                    self._register[key]['onchange'](getattr(self, key))

    def __restrict_keys(self, d, keys):
        if keys is None:
            return d
        return {key: d[key] for key in keys if key in d}

    def save(self, filename=None, keys=None, replace=None):
        '''
        Save the current configuration as a JSON file. Accepted arguments are:
         - filename : The location of the file to be saved. If not specified,
            default configuration location is used (and autoloaded on startup).
         - keys : The keys to be saved. If `None`, all keys are saved (or set to
            default values if missing).
         - replace : Whether the configuration file should be replaced (True), or
            simply updated (False). If False, then the existing keys stored in the
            maintained except where they conflict with the keys specified. The
            default value is `None`, in which case it maps to `True` if keys=None,
            or `False` if specific keys are specified. (default=None)
        '''
        filename = filename or self._config_path
        filename = os.path.join(ensure_path_exists(os.path.dirname(filename)), os.path.basename(filename))
        config = {}
        if replace is None:
            replace = True if keys is None else False
        if keys is None:
            replace = True
        if not replace and os.path.exists(filename):
            with io.open(filename, 'r') as f:
                config = json.load(f)
        config.update(self.__restrict_keys(self._config, keys))
        with io.open(filename, 'w') as f:
            json_config = json.dumps(config, ensure_ascii=False, indent=4)
            if sys.version_info.major == 2 and isinstance(json_config, six.string_types):
                json_config = json_config.decode("utf-8")
            f.write(json_config)

    def load(self, filename=None, keys=None, replace=None, force=False):
        '''
        Load a configuration from the disk. Accepted arguments are:
         - filename : The location of the configuration. By default, this will
            point to the automatically loaded configuration file.
         - keys : The keys to load from the configuration. If `None`, all keys
            are loaded from config file (or set to default values if missing).
         - replace : Whether the current configuration should be replaced (True), or
           simply updated (False). If False, then the existing configuration will be
           maintained except where conflicts exist with the keys being loaded. The
           default value is `None`, in which case it maps to `True` if keys=None,
           or `False` if specific keys are specified. (default=None)
         - force : Ordinarily, new configuration is run through the standard checks
            but in some cases (such as startup), the register has yet to be filled,
            and so results in spouts spurious warnings. This allows one to bypass
            all checks.
        '''
        filename = filename or self._config_path
        if replace is None:
            replace = True if keys is None else False
        if keys is None:
            replace = True
        with io.open(filename, 'r') as f:
            config = self.__restrict_keys(json.load(f), keys)
            if force:
                self._config = config
            else:
                if replace:
                    self.reset(**config)
                else:
                    self.reset(*keys, **config)


config = Configuration()



import datetime
import getpass
import os
import re
import tempfile

import pandas as pd
from omniduct.remotes.base import RemoteClient
from omniduct.utils.debug import logger
from omniduct.utils.ports import is_local_port_free
from omniduct.utils.processes import run_in_subprocess

from omniduct.errors import DuctAuthenticationError

SSH_ASKPASS = '{omniduct_dir}/utils/ssh_askpass'.format(omniduct_dir=os.path.dirname(__file__))
SESSION_SSH_USERNAME = None
SESSION_REMOTE_HOST = None
SESSION_SSH_ASKPASS = False


class SSHClient(RemoteClient):
    """SSHClient manages a persistent connection to remote hosts allowing
    remote execution, port forwarding, file transfer, and cleanup

    References
    ----------
    https://puppetlabs.com/blog/speed-up-ssh-by-reusing-connections

    """

    PROTOCOLS = ['ssh']

    def _init(self):
        pass

    # Duct connection implementation

    def _connect(self):
        """
        Create persistent connection to remote host.

        The workflow to handle passwords and host keys is inspired by the pxssh module of pexpect
        (https://github.com/pexpect/pexpect). We have adjusted this workflow to our purposes.

        Returns
        -------
        proc : Popen subprocess
            Subprocess used to connect.
        """
        import pexpect

        # Create socket directory if it doesn't exist.
        socket_dir = os.path.dirname(self._socket_path)
        if not os.path.exists(socket_dir):
            os.makedirs(socket_dir)
        # Create persistent master connection and exit.
        cmd = ("ssh {login} -MT "
               "-S {socket} "
               "-o ControlPersist=yes "
               "-o StrictHostKeyChecking=no "
               "-o NoHostAuthenticationForLocalhost=yes "
               "-o ServerAliveInterval=60 "
               "-o ServerAliveCountMax=2 "
               "'exit'".format(login=self._login_info, socket=self._socket_path))

        expected = [
            "WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!",    # 0
            "(?i)are you sure you want to continue connecting",    # 1
            "(?i)(?:(?:password)|(?:passphrase for key)):",        # 2
            "(?i)permission denied",                               # 3
            "(?i)terminal type",                                   # 4
            pexpect.TIMEOUT,                                       # 5
            "(?i)connection closed by remote host",                # 6
            "(?i)could not resolve hostname",                      # 7
            pexpect.EOF                                            # 8
        ]

        try:
            expect = pexpect.spawn(cmd)
            i = expect.expect(expected, timeout=10)

            # First phase
            if i == 0:  # If host identification changed, arrest any further attempts to connect
                raise RuntimeError('Host identification for {} has changed! If you are sure that the host identification SHOULD have changed, and are not experiencing a man-in-the-middle attack, please remove the line corresponding to this server in ~/.ssh/known_hosts; or call the `update_host_keys` method of this object.'.format(self._host))
            if i == 1:  # Request to authorize host certificate (i.e. host not in the 'known_hosts' file)
                expect.sendline("yes")
                i = self.expect(expected)
            if i == 2:  # Request for password/passphrase
                expect.sendline(self.password or getpass.getpass('Password: '))
                i = self.expect(expected)
            if i == 4:  # Request for terminal type
                expect.sendline('ascii')
                i = self.expect(expected)

            # Second phase
            if i == 1:  # Another request to authorize host certificate (i.e. host not in the 'known_hosts' file)
                raise RuntimeError('Received a second request to authorize host key. This should not have happened!')
            elif i in (2, 3):  # Second request for password/passphrase or rejection of credentials. For now, give up.
                raise DuctAuthenticationError('Invalid username and/or password, or private key is not unlocked.')
            elif i == 4:  # Another request for terminal type.
                raise RuntimeError('Received a second request for terminal type. This should not have happened!')
            elif i == 5:  # Timeout
                # In our instance, this means that we have not handled some or another aspect of the login procedure.
                # Since we are expecting an EOF when we have successfully logged in, hanging means that the SSH login
                # procedure is waiting for more information. Since we have no more to give, this means our login
                # was unsuccessful.
                raise RuntimeError('SSH client seems to be awaiting more information, but we have no more to give. The '
                                   'messages received so far are:\n{}'.format(expect.before))
            elif i == 6:  # Connection closed by remote host
                raise RuntimeError("Remote closed SSH connection")
            elif i == 7:
                raise RuntimeError("Cannot connect to {} on your curent network connection".format(self.host))
        finally:
            expect.close()

        # We should be logged in at this point, but let us make doubly sure
        assert self.is_connected(), 'Unexpected failure to establish a connection with the remote host with command: \n ' \
                                    '{}\n\n Please report this!'.format(cmd)

    def _is_connected(self):
        """
        Return whether SSHClient is connected by checking the control socket.
        """
        cmd = "ssh {login} -T -S {socket} -O check".format(login=self._login_info,
                                                           socket=self._socket_path)
        proc = run_in_subprocess(cmd)

        if proc.returncode != 0:
            if os.path.exists(self._socket_path):
                os.remove(self._socket_path)
            return False
        return True

    def _disconnect(self):
        """
        Exit persistent connection to remote host.
        """
        # Send exit request to control socket.
        cmd = "ssh {login} -T -S {socket} -O exit".format(login=self._login_info,
                                                          socket=self._socket_path)
        proc = run_in_subprocess(cmd)

    # RemoteClient implementation

    def _execute(self, cmd, **kwargs):
        """
        Execute a command on a remote host.

        Parameters
        ----------
        cmd : string
            Command to be executed on remote host.
        kwargs : keywords
            Options to pass to subprocess.Popen.

        Returns
        -------
        proc : Popen subprocess
            Subprocess used to run remote command.
        """
        template = 'ssh {login} -T -o ControlPath={socket} << EOF\n{cmd}\nEOF'
        config = dict(self._subprocess_config)
        config.update(kwargs)
        return run_in_subprocess(template.format(login=self._login_info,
                                                 socket=self._socket_path,
                                                 cmd=cmd),
                                 check_output=True,
                                 **config)

    def _port_forward_start(self, local_port, remote_host, remote_port):
        self.connect()
        logger.info('Establishing port forward...')
        cmd_template = 'ssh {login} -T -O forward -S {socket} -L localhost:{local_port}:{remote_host}:{remote_port}'
        cmd = cmd_template.format(login=self._login_info,
                                  socket=self._socket_path,
                                  local_port=local_port,
                                  remote_host=remote_host,
                                  remote_port=remote_port)
        proc = run_in_subprocess(cmd)
        if proc.returncode != 0:
            raise Exception('Unable to port forward with command: {}'.format(cmd))
        logger.info(proc.stderr or 'Success')
        return proc

    def _port_forward_stop(self, local_port, remote_host, remote_port, connection):
        logger.info('Cancelling port forward...')
        cmd_template = 'ssh {login} -T -O cancel -S {socket} -L localhost:{local_port}:{remote_host}:{remote_port}'
        cmd = cmd_template.format(login=self._login_info,
                                  socket=self._socket_path,
                                  local_port=local_port,
                                  remote_host=remote_host,
                                  remote_port=remote_port)
        proc = run_in_subprocess(cmd)
        logger.info('Success' if proc.returncode == 0 else 'Failure')

    def _is_port_bound(self, host, port):
        return self.execute('which nc; if [ $? -eq 0 ]; then nc -z -w2 {} {}; fi'.format(host, port)).returncode == 0

    # FileSystem methods

    def _exists(self, path):
        return self.execute('if [ ! -e {} ]; then exit 1; fi'.format(path)).returncode == 0

    def _isdir(self, path):
        return self.execute('if [ ! -d {} ]; then exit 1; fi'.format(path)).returncode == 0

    def _isfile(self, path):
        return self.execute('if [ ! -f {} ]; then exit 1; fi'.format(path)).returncode == 0

    def _listdir(self, path):
        path = os.path.join('~', path or '')
        return sorted([f for f in self.execute('ls -a {}'.format(path)).stdout.decode().strip().split('\n') if f not in ['.', '..']])

    def _showdir(self, path):
        path = os.path.join('~', path or '')
        dir = pd.DataFrame(sorted([re.split('\s+', f) for f in self.execute('ls -al {}'.format(path)).stdout.decode().strip().split('\n')[1:]]),
                           columns=['file_mode', 'link_count', 'owner', 'group', 'bytes', 'month', 'day', 'time', 'path'])

        def convert_to_datetime(x):
            months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
            year = datetime.datetime.now().year if ':' in x.time else x.time
            time = x.time if ':' in x.time else None
            return datetime.datetime(
                year=int(year),
                month=months.index(x.month) + 1,
                day=int(x.day),
                hour=int(time.split(':')[0]) if time is not None else 0,
                minute=int(time.split(':')[1]) if time is not None else 0
            )

        return dir.assign(
            last_modified=lambda x: x.apply(convert_to_datetime, axis=1),
            type=lambda x: x.apply(lambda x: 'directory' if x.file_mode.startswith('d') else 'file', axis=1)
        ).drop(
            ['month', 'day', 'time'],
            axis=1
        ).sort_values(
            ['type', 'path']
        ).reset_index(drop=True)[['type', 'path', 'bytes', 'last_modified', 'file_mode', 'owner', 'group', 'link_count']]

    # File handling

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        read = self.execute('cat {}'.format(path)).stdout
        if not binary:
            read = read.decode()
        return read

    def _file_append_(self, path, s, binary):
        raise NotImplementedError

    def _file_write_(self, path, s, binary):
        if binary:
            fd, tmp_path = tempfile.mkstemp()
        else:
            fd, tmp_path = tempfile.mkstemp(text=True)
        os.close(fd)

        with open(tmp_path, 'w' + ('b' if binary else '')) as f:
            f.write(s)

        return self._copy_from_local(tmp_path, path, overwrite=True)

    # File transfer

    def _copy_to_local(self, source, dest, overwrite):
        """
        SCP remote file.

        Parameters
        ----------
        origin_file : string
            Path to local file to copy.
        destination_file : string
            Target location on remote host.
        """
        self.connect()
        logger.info('Copying file to local...')
        template = 'scp -o ControlPath={socket} {login}:{remote_file} {local_file}'
        destination_file = dest or os.path.split(source)[1]
        proc = run_in_subprocess(template.format(socket=self._socket_path,
                                                 login=self._login_info,
                                                 local_file=destination_file,
                                                 remote_file=source),
                                 check_output=True)
        logger.info(proc.stderr or 'Success')

    def _copy_from_local(self, source, dest, overwrite):
        """
        SCP local file.

        Parameters
        ----------
        origin_file : string
            Path to remote file to copy.
        destination_file : string
            Target location on local machine.
        """
        self.connect()
        logger.info('Copying file from local...')
        template = 'scp -o ControlPath={socket} {local_file} {login}:{remote_file}'
        destination_file = dest or os.path.split(source)[1]
        proc = run_in_subprocess(template.format(socket=self._socket_path,
                                                 login=self._login_info,
                                                 local_file=source,
                                                 remote_file=destination_file),
                                 check_output=True)
        logger.info(proc.stderr or 'Success')

    # Helper methods

    @property
    def _login_info(self):
        return '@'.join([self.username, self.host])

    @property
    def _socket_path(self):
        # On Linux the maximum socket path length is 108 characters, and on Mac OS X it is 104 characters, including
        # the final sentinel character (or so it seems). SSH appends a '.' character, followed by random sequence of 16
        # characters. We therefore need the rest of the path to be less than 86 characters.
        return os.path.expanduser('~/.ssh/omniduct/{}'.format(self._login_info))[:86]

    @property
    def _subprocess_config(self):
        return {}

    def update_host_keys(self):
        assert not self.remote, "Updating host key only works for local connections."
        cmd = "ssh-keygen -R {host} && ssh-keyscan {host} >> ~/.ssh/known_hosts".format(host=self.host)
        return run_in_subprocess(cmd, False).returncode == 0
